import app/web
import gleam/bit_array
import gleam/bytes_tree
import gleam/crypto
import gleam/dynamic/decode
import gleam/erlang/process.{type Subject}
import gleam/float
import gleam/http/request.{type Request}
import gleam/http/response.{type Response}
import gleam/int
import gleam/io
import gleam/json
import gleam/list
import gleam/option.{type Option, None, Some}
import gleam/otp/actor
import gleam/result
import gleam/string
import mist.{type Connection, type ResponseData}
import player
import player_session
import session
import socket_manager
import storail

pub fn handle_request(
  req: Request(Connection),
  ctx: web.Context,
) -> Response(ResponseData) {
  io.println("Got request: " <> req.path)
  case request.path_segments(req) {
    [] -> serve_wasm(ctx, "index.html", "text/html")
    ["static", "frontend.js"] ->
      serve_wasm(ctx, "frontend.js", "text/javascript")
    ["static", "frontend_bg.wasm"] ->
      serve_wasm(ctx, "frontend_bg.wasm", "application/wasm")
    ["ws"] -> serve_websocket(ctx, req)
    _ -> response.set_body(response.new(404), mist.Bytes(bytes_tree.new()))
  }
}

fn serve_wasm(
  ctx: web.Context,
  file: String,
  mimetype: String,
) -> Response(ResponseData) {
  let wasm_path = string.join([ctx.config.static_path, file], "/")
  mist.send_file(wasm_path, offset: 0, limit: None)
  |> result.map(fn(file) {
    response.new(200)
    |> response.prepend_header("content-type", mimetype)
    |> response.set_body(file)
  })
  |> result.lazy_unwrap(fn() {
    response.new(404)
    |> response.set_body(mist.Bytes(bytes_tree.new()))
  })
}

pub type SocketState {
  SocketState(
    ctx: web.Context,
    id: Int,
    subject: Subject(web.GameResponse),
    user_id: Option(Int),
    game_id: Option(String),
  )
}

fn serve_websocket(
  ctx: web.Context,
  req: Request(Connection),
) -> Response(ResponseData) {
  mist.websocket(
    request: req,
    on_init: fn(_conn) {
      let self = process.new_subject()

      let selector =
        process.new_selector()
        |> process.selecting(self, fn(x) { x })

      let id = socket_manager.add_socket_subj(ctx.socket_manager, self)

      let state =
        SocketState(ctx:, id:, subject: self, user_id: None, game_id: None)
      #(state, Some(selector))
    },
    on_close: fn(state) {
      io.println("Shutting down socket" <> int.to_string(state.id))
      socket_manager.remove_socket_subj(state.ctx.socket_manager, state.subject)
    },
    handler: handle_ws_message,
  )
}

fn decode_error_to_string(decode_err: decode.DecodeError) -> String {
  "Path: "
  <> string.join(decode_err.path, ".")
  <> " Expected: "
  <> decode_err.expected
  <> " Got: "
  <> decode_err.found
}

fn decode_errors_to_string(decode_errs: List(decode.DecodeError)) -> String {
  string.join(list.map(decode_errs, decode_error_to_string), "\n")
}

fn handle_ws_message(
  state: SocketState,
  conn,
  message: mist.WebsocketMessage(web.GameResponse),
) {
  io.println("Got WS message")
  case message {
    mist.Text("ping") -> {
      io.println("Got Ping")
      let assert Ok(_) = mist.send_text_frame(conn, "pong")
      actor.continue(state)
    }
    mist.Text(msg) -> {
      echo msg
      let state = case json.parse(msg, web.client_request_decoder()) {
        Error(err) -> {
          let err_msg = case err {
            json.UnableToDecode(decode_err) ->
              "Decode Error: " <> decode_errors_to_string(decode_err)
            json.UnexpectedByte(bad_byte) -> "Unexpeted byte: " <> bad_byte
            json.UnexpectedEndOfInput -> "Reached EoF"
            json.UnexpectedFormat(_) -> "Unexpected format"
            json.UnexpectedSequence(seq_msg) ->
              "Unexpected Sequence: " <> seq_msg
          }
          io.println("Could not parse client message: " <> err_msg)
          state
        }
        Ok(req) -> {
          let #(state, resp) = handle_client_msg(state, req)
          echo resp
          let resp_str =
            web.encode_game_response(resp)
            |> json.to_string

          let assert Ok(_) = mist.send_text_frame(conn, resp_str)
          state
        }
      }

      actor.continue(state)
    }
    mist.Binary(_) -> {
      actor.continue(state)
    }
    mist.Custom(resp) -> {
      echo resp
      let resp_str =
        web.encode_game_response(resp)
        |> json.to_string
      let assert Ok(_) = mist.send_text_frame(conn, resp_str)
      actor.continue(state)
    }
    mist.Closed | mist.Shutdown -> actor.Stop(process.Normal)
  }
}

fn register_user(
  state: SocketState,
  username: String,
  gamecode: String,
) -> #(String, Int) {
  let secret = crypto.strong_random_bytes(16)
  let token =
    crypto.new_hasher(crypto.Sha512)
    |> crypto.hash_chunk(secret)
    |> crypto.hash_chunk(bit_array.from_string(username))
    |> crypto.hash_chunk(bit_array.from_string(gamecode))
    |> crypto.digest
    |> bit_array.base64_encode(True)

  let token_hash = player_session.hash_token(token)

  let assert <<user_id:int-size(64)>> = crypto.strong_random_bytes(8)

  let key = storail.key(state.ctx.player_sessions, int.to_base16(user_id))

  let assert Ok(Nil) =
    storail.write(
      key,
      player_session.PlayerSession(
        id: user_id,
        token_hash:,
        username:,
        socket_id: state.id,
      ),
    )

  #(token, user_id)
}

fn handle_client_msg(
  state: SocketState,
  req: web.ClientRequest,
) -> #(SocketState, web.GameResponse) {
  case req.msg {
    web.BuzzIn(time) -> {
      io.println("Got buzz in @ " <> float.to_string(time))
      #(state, web.AckBuzzer)
    }
    web.Register(gamecode, username) -> {
      let #(token, user_id) = register_user(state, username, gamecode)
      let key = storail.key(state.ctx.sessions, gamecode)
      let assert Ok(session) = storail.read(key)

      let player = player.Player(name: username, id: user_id, score: 0)
      let session =
        session.Session(
          ..session,
          players: list.append(session.players, [player]),
        )

      let assert Ok(Nil) = storail.write(key, session)

      #(
        SocketState(..state, user_id: Some(user_id), game_id: Some(gamecode)),
        web.JoinResponse(username, token),
      )
    }
    web.CreateRoom(gamecode, username) -> {
      let #(token, user_id) = register_user(state, username, gamecode)

      let key = storail.key(state.ctx.sessions, gamecode)
      let session = session.Session(gamecode, user_id, [])
      let assert Ok(Nil) = storail.write(key, session)

      #(
        SocketState(..state, user_id: Some(user_id), game_id: Some(gamecode)),
        web.HostResponse(username, token),
      )
    }
    web.ResetBuzzers -> {
      let _token_hash = player_session.hash_token(req.token)
      let assert Some(user_id) = state.user_id
      let assert Some(session_id) = state.game_id

      let player_key =
        storail.key(state.ctx.player_sessions, int.to_base16(user_id))

      let assert Ok(_user) = storail.read(player_key)

      let session_key = storail.key(state.ctx.sessions, session_id)
      let assert Ok(session) = storail.read(session_key)

      list.map(session.players, fn(player) { player.id })
      |> list.map(fn(player_id) {
        let player_key =
          storail.key(state.ctx.player_sessions, int.to_base16(player_id))
        let assert Ok(player) = storail.read(player_key)
        player
      })
      |> list.map(fn(player) {
        let assert Ok(subject) =
          socket_manager.get_socket_subj_by_id(
            state.ctx.socket_manager,
            player.socket_id,
          )
        subject
      })
      |> list.each(fn(subject) { actor.send(subject, web.ResetBuzzer) })

      #(state, web.AckBuzzer)
    }
  }
}
